"""constants and other definitions"""
import struct

ID_GET = "\x01"
ID_SET = "\x02"
ID_DEL = "\x03"
ID_GETKEYS = "\x04"
ID_ANSWER = "\x05"
ID_NOTFOUND = "\x06"
ID_ALLKEYS = "\x07"
ID_SWITCH = "\x08"

VERSION = 1
VERSION_FORMAT = "!Q"
VERSION_LENGTH = struct.calcsize(VERSION_FORMAT)

MODE_VERSION = "\x01"
MODE_PASSWORD = "\x02"
MODE_UNKNOWN = "\x03"
MODE_SERVER = "\x04"
MODE_CLIENT = "\x05"
MODE_ERROR = "\x06"
MODE_CONNECTING = "\x07"
MODE_RANGE = "\x08"

STATE_OK = "O"
STATE_ERROR = "E"
STATE_PASSWORD_REQUIRED = "P"

MESSAGE_LENGTH_FORMAT = "!Q"
MESSAGE_LENGTH_SIZE = struct.calcsize(MESSAGE_LENGTH_FORMAT)
MESSAGE_KEY_LENGTH_FORMAT = "!L"
MESSAGE_KEY_LENGTH_SIZE = struct.calcsize(MESSAGE_KEY_LENGTH_FORMAT)
MESSAGE_ID_FORMAT = "!I"
MESSAGE_ID_SIZE = struct.calcsize(MESSAGE_ID_FORMAT)

DEFAULT_PORT = 54565

INVALID_PASSWORD_SLEEP_INTERVAL = 3

RANGE = 2**20
RANGE_FORMAT = "!QQ"
RANGE_SIZE = struct.calcsize(RANGE_FORMAT)



"""dirdbm database interface."""
from twisted.persisted import dirdbm


class DirdbmDatabase(object):
    """A dirdbm database"""
    def __init__(self, args):
        self.args = args
        if len(self.args) != 1:
            raise ValueError, "Expected exactly one argument for the DB."
        self.path = args[0]
        self.db = dirdbm.DirDBM(self.path)

    def get(self, key):
        """returns the value for key."""
        if key in self.db:
            return self.db[key]
        else:
            raise KeyError(key)

    def set(self, key, value):
        """sets key to value"""
        self.db[key] = value

    def delete(self, key):
        """deletes the key/value pair for key."""
        try:
            del self.db[key]
        except KeyError:
            pass

    def getkeys(self):
        """returns a list of keys."""
        return self.db.keys()

    def reset(self):
        """resets the database."""
        self.db.clear()

    def close(self):
        """this should close the database."""
        self.db.close()  # this is no-op



"""utility functions"""
import struct

from twisted.internet.defer import Deferred

from . import data


def keystring2list(s):
    """convert a string of keys to a list of keys."""
    if len(s) == 0:
        return []
    keys = []
    i = 0
    while i < len(s):
        keylength = struct.unpack(data.MESSAGE_KEY_LENGTH_FORMAT, s[i:i + data.MESSAGE_KEY_LENGTH_SIZE])[0]
        i += data.MESSAGE_KEY_LENGTH_SIZE
        key = s[i:i + keylength]
        keys.append(key)
        i += keylength
    return keys


def keylist2string(keys):
    """converts a list of keys to a string."""
    answer = []
    for key in keys:
        keylength = len(key)
        lengthdata = struct.pack(data.MESSAGE_KEY_LENGTH_FORMAT, keylength)
        answer += [lengthdata, str(key)]
    return "".join(answer)


def dsleep(reactor, s):
    """return a deferred, which will be called after s seconds."""
    d = Deferred()
    reactor.callLater(s, d.callback, None)
    return d


def fmtcols(mylist, cols):
    """
formats a list into columns.
Copied from https://stackoverflow.com/questions/171662/formatting-a-list-of-text-into-columns
"""
    lines = ("\t".join(mylist[i:i + cols]) for i in xrange(0, len(mylist), cols))
    return '\n'.join(lines)



"""asynchronous clientside protocol for twisted."""
import struct

from twisted.protocols.basic import IntNStringReceiver
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.python import log

from . import data, utils


class VersionMismatch(Exception):
    """Version does not match"""
    pass


class ProtocolError(Exception):
    """invalid answer/message"""
    pass


class IncorrectPassword(ValueError):
    """Invalid password."""
    pass


class ClientProtocol(IntNStringReceiver):
    """
This protocol connects to the router.
You can use the 'callback' attribute to get a deferred which will be called once the handshake is finished.
"""
    def __init__(self, password):
        if hasattr(IntNStringReceiver, "__init__"):
            IntNStringReceiver.__init__(self)
        self.password = password
        self.requests = {}
        self.free = set()
        self.range_start = None
        self.range_end = None
        self.mode = data.MODE_CONNECTING
        self.structFormat = data.MESSAGE_LENGTH_FORMAT
        self.prefixLength = data.MESSAGE_LENGTH_SIZE
        self.callback = Deferred()  # will be called once the handshake is finished

    def connectionMade(self):
        """called when the connection was established."""
        if hasattr(IntNStringReceiver, "connectionMade"):
            IntNStringReceiver.connectionMade(self)
        # init handshake
        log.msg("Connection established, initiating handshake...")
        self.mode = data.MODE_VERSION
        self.sendString(struct.pack(data.VERSION_FORMAT, data.VERSION))

    def stringReceived(self, msg):
        """called when a message was received."""
        if self.mode == data.MODE_ERROR:
            # error occured, connection is probably already being terminated.
            log.err("Received a message, but the protocol is already mismatching. Ignoring Message.")
            pass

        elif self.mode == data.MODE_VERSION:
            # check version answer
            if msg == data.STATE_OK:
                # Version OK, continue
                log.msg("Version OK, sending mode and requesting range...")
                self.sendString(data.MODE_CLIENT)
                self.mode = data.MODE_RANGE
            elif msg == data.STATE_ERROR:
                # Version mismatch
                log.err("Version is not supported by the server. Losing Connection...")
                self.mode = data.MODE_ERROR
                f = Failure(VersionMismatch("Version does not match!"))
                try:
                    self.transport.loseConnection()
                except:
                    pass
                finally:
                    self.callback.errback(f)
            elif msg == data.STATE_PASSWORD_REQUIRED:
                # Version OK, but password required
                if self.password is not None:
                    log.msg("Version OK, sending password...")
                    self.sendString(self.password)
                    self.mode = data.MODE_PASSWORD
                else:
                    log.err("Version OK, but a password is required!")
                    self.mode = data.MODE_ERROR
                    self.transport.loseConnection()
                    self.callback.errback(Failure(IncorrectPassword("Password required, but None specified!")))
            else:
                # unknown answer
                log.err("Received unknown answer! Losing Connection...")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(ProtocolError("Unknown Answer during version check!")))
                self.transport.loseConnection()

        elif self.mode == data.MODE_PASSWORD:
            # check answer for password
            if msg == data.STATE_OK:
                # Password OK, continue
                log.msg("Password OK, sending mode and requesting range...")
                self.mode = data.MODE_RANGE
                self.sendString(data.MODE_CLIENT)  # < -- this is correct
            elif msg == data.STATE_ERROR:
                # Password invalid
                log.err("Invalid password!")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(IncorrectPassword("Invalid Password!")))
            else:
                # invalid answer
                log.err("Received unknown Answer. Losing Connection...")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(ProtocolError("Unknown Answer during password check!")))
                self.transport.loseConnection()

        elif self.mode == data.MODE_RANGE:
            self.range_start, self.range_end = struct.unpack(data.RANGE_FORMAT, msg)
            log.msg("Range is {s} to {e}.".format(s=self.range_start, e=self.range_end))
            self.cur_id = self.range_start
            self.mode = data.MODE_CLIENT
            self.callback.callback(self)

        elif self.mode == data.MODE_CLIENT:
            # handle answers
            actionbyte = msg[0]
            msg = msg[1:]
            if actionbyte in (data.ID_ALLKEYS, data.ID_ANSWER, data.ID_NOTFOUND):
                rids = msg[:data.MESSAGE_ID_SIZE]
                rid = struct.unpack(data.MESSAGE_ID_FORMAT, rids)[0]
                self.free.add(rid)
                if rid not in self.requests:
                    pass
                else:
                    d = self.requests[rid]
                    del self.requests[rid]
                    if actionbyte == data.ID_ANSWER:
                        value = msg[data.MESSAGE_ID_SIZE:]
                        d.callback(value)
                    elif actionbyte == data.ID_NOTFOUND:
                        f = Failure(KeyError("Key not found!"))
                        d.errback(f)
                    elif actionbyte == data.ID_ALLKEYS:
                        keystring = msg[data.MESSAGE_ID_SIZE:]
                        keylist = utils.keystring2list(keystring)
                        d.callback(keylist)
                    else:
                        log.err("Logic Error! Expected all answer IDs to have been checked!")
            else:
                log.err("Error: Unexpected Answer from server! Losing Connection...")
                self.mode = data.MODE_ERROR
                self.transport.loseConnection()
                self.callback.errback(Failure(ProtocolError("Unknown Answer!")))

        else:
            log.err("Logic Error: set protocol to unknown mode!")

    def get_id(self):
        """returns a request id."""
        if len(self.free) > 0:
            return self.free.pop()
        else:
            i = self.cur_id
            if self.cur_id + 1 >= self.range_end:
                raise RuntimeError("No ids left.")
            self.cur_id += 1
            return i

    def get(self, key):
        """returns a deferred which will be fired with the value of key."""
        assert isinstance(key, str), "Expected key to be a string!"
        d = Deferred()
        rid = self.get_id()
        rids = struct.pack(data.MESSAGE_ID_FORMAT, rid)
        self.requests[rid] = d
        self.sendString(data.ID_GET + rids + key)
        return d

    def set(self, key, value):
        """sets key to value."""
        assert isinstance(key, str) and isinstance(value, str), "Expected key/value to be strings!"
        keylength = len(key)
        keylengthstring = struct.pack(data.MESSAGE_KEY_LENGTH_FORMAT, keylength)
        self.sendString(data.ID_SET + keylengthstring + key + value)

    def delete(self, key):
        """deletes the value for key and key."""
        assert isinstance(key, str), "Expected key to be a string!"
        self.sendString(data.ID_DEL + key)

    def getkeys(self):
        """returns a deferred which will be fired with a list of all keys"""
        d = Deferred()
        rid = self.get_id()
        rids = struct.pack(data.MESSAGE_ID_FORMAT, rid)
        self.requests[rid] = d
        self.sendString(data.ID_GETKEYS + rids)
        return d



"""Databaseserver side protocol of the routing."""
import struct
import os

from twisted.protocols.basic import IntNStringReceiver
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python.failure import Failure
from twisted.python import log

from . import data, utils


class VersionMismatch(Exception):
    """Version does not match"""
    pass


class ProtocolError(Exception):
    """invalid answer/message"""
    pass


class IncorrectPassword(ValueError):
    """Invalid password."""
    pass


class DatabaseClientProtocol(IntNStringReceiver):
    """This protocol connects a database to the router."""
    def __init__(self, db, password, reactor, reset=False, reset_sleep_interval=0.2):
        if hasattr(IntNStringReceiver, "__init__"):
            IntNStringReceiver.__init__(self)
        self.db = db
        self.reactor = reactor
        self.password = password
        self.reset = reset
        self.sleep_interval = reset_sleep_interval
        self.reset_req_string = os.urandom(data.MESSAGE_ID_SIZE)
        self.to_sync = []
        self.reset_requests = {}
        self.range_start = None
        self.range_end = None
        self.free = set()
        self.cur_id = 0
        self.mode = data.MODE_CONNECTING
        self.structFormat = data.MESSAGE_LENGTH_FORMAT
        self.prefixLength = data.MESSAGE_LENGTH_SIZE
        self.callback = Deferred()  # will be called once the handshake is finished

    def connectionMade(self):
        """called when the connection was established."""
        if hasattr(IntNStringReceiver, "connectionMade"):
            IntNStringReceiver.connectionMade(self)
        # init handshake
        log.msg("Connection established, initiating handshake...")
        self.mode = data.MODE_VERSION
        self.sendString(struct.pack(data.VERSION_FORMAT, data.VERSION))

    def connectionLost(self, reason=None):
        """this will be called when the connection was lost."""
        log.err("Error: Connection lost. Reason: {r}".format(r=reason))
        log.msg("Closing database...")
        self.db.close()
        log.msg("Database closed.")

    @inlineCallbacks
    def stringReceived(self, msg):
        """called when a message was received."""
        if self.mode == data.MODE_ERROR:
            # error occured, connection is probably already being terminated.
            log.err("Received a message, but the protocol is already mismatching. Ignoring Message.")
            pass

        elif self.mode == data.MODE_VERSION:
            # check version answer
            if msg == data.STATE_OK:
                # Version OK, continue
                log.msg("Version OK, sending mode and requesting range...")
                self.sendString(data.MODE_SERVER)
                self.mode = data.MODE_RANGE  # <--- this is correct
            elif msg == data.STATE_ERROR:
                # Version mismatch
                log.err("Version is not supported by the server. Losing Connection...")
                self.mode = data.MODE_ERROR
                f = Failure(VersionMismatch("Version does not match!"))
                try:
                    self.transport.loseConnection()
                except:
                    pass
                finally:
                    self.callback.errback(f)
            elif msg == data.STATE_PASSWORD_REQUIRED:
                # Version OK, but password required
                if self.password is not None:
                    log.msg("Version OK, sending password...")
                    self.sendString(self.password)
                    self.mode = data.MODE_PASSWORD
                else:
                    log.err("Version OK, but a password is required!")
                    self.mode = data.MODE_ERROR
                    self.transport.loseConnection()
                    self.callback.errback(Failure(IncorrectPassword("Password required, but None specified!")))
            else:
                # unknown answer
                log.err("Received unknown answer! Losing Connection...")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(ProtocolError("Unknown Answer during version check!")))
                self.transport.loseConnection()

        elif self.mode == data.MODE_PASSWORD:
            # check answer for password
            if msg == data.STATE_OK:
                # Password OK, continue
                log.msg("Password OK, sending mode and requesting range...")
                self.mode = data.MODE_RANGE
                self.sendString(data.MODE_SERVER)  # <--- this is correct
            elif msg == data.STATE_ERROR:
                # Password invalid
                log.err("Invalid password!")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(IncorrectPassword("Invalid Password!")))
            else:
                # invalid answer
                log.err("Received unknown Answer. Losing Connection...")
                self.mode = data.MODE_ERROR
                self.callback.errback(Failure(ProtocolError("Unknown Answer during password check!")))
                self.transport.loseConnection()

        elif self.mode == data.MODE_RANGE:
            self.range_start, self.range_end = struct.unpack(data.RANGE_FORMAT, msg)
            log.msg("Range is {s} to {e}.".format(s=self.range_start, e=self.range_end))
            self.cur_id = self.range_start
            self.mode = data.MODE_SERVER
            if self.reset:
                self.init_reset()
            else:
                self.callback.callback(self)

        elif self.mode == data.MODE_SERVER:
            # db requests.
            actionbyte = msg[0]
            msg = msg[1:]
            if actionbyte == data.ID_SET:
                keylength = struct.unpack(data.MESSAGE_KEY_LENGTH_FORMAT, msg[:data.MESSAGE_KEY_LENGTH_SIZE])[0]
                key = msg[data.MESSAGE_KEY_LENGTH_SIZE:data.MESSAGE_KEY_LENGTH_SIZE + keylength]
                if self.reset:
                    if key in self.to_sync:
                        self.to_sync.remove(key)
                value = msg[data.MESSAGE_KEY_LENGTH_SIZE + keylength:]
                self.db.set(key, value)
            elif actionbyte == data.ID_GET:
                rid = msg[:data.MESSAGE_ID_SIZE]
                key = msg[data.MESSAGE_ID_SIZE:]
                try:
                    value = yield self.db.get(key)
                except KeyError:
                    self.sendString(data.ID_NOTFOUND + rid)
                else:
                    self.sendString(data.ID_ANSWER + rid + str(value))

            elif actionbyte == data.ID_DEL:
                key = msg
                if key in self.to_sync:
                    self.to_sync.remove(key)
                try:
                    self.db.delete(key)
                except:
                    pass

            elif actionbyte == data.ID_GETKEYS:
                rid = msg
                keys = yield self.db.getkeys()
                answer = utils.keylist2string(keys)
                self.sendString(data.ID_ALLKEYS + rid + answer)

            elif actionbyte == data.ID_ALLKEYS:
                # answer during reset
                rid = msg[:data.MESSAGE_ID_SIZE]
                self.free.add(struct.unpack(data.MESSAGE_ID_FORMAT, rid)[0])
                if rid == self.reset_req_string:
                    log.msg("Received keylist, starting value requests...")
                    keystring = msg[data.MESSAGE_ID_SIZE:]
                    keys = utils.keystring2list(keystring)
                    if len(keys) == 0:
                        log.msg("Keylist empty; skipping value requests...")
                        self.sendString(data.ID_SWITCH)
                        self.callback.callback(self)
                        pass
                    else:
                        # request all values for key
                        self.to_sync = keys
                        del keys  # this list may be big, we should delete it to free some RAM
                        done = 0
                        while True:
                            if len(self.to_sync) == 0:
                                break
                            rid = self.get_id()
                            key = self.to_sync.pop(0)
                            self.reset_requests[rid] = key
                            rids = struct.pack(data.MESSAGE_ID_FORMAT, rid)
                            self.sendString(data.ID_GET + rids + key)
                            if (done % 128) == 0:
                                yield utils.dsleep(self.reactor, self.sleep_interval)
                            done += 1
                        log.msg("Finished sending value requests.")
                else:
                    log.err("Received invalid RID during sync!")

            elif actionbyte == data.ID_ANSWER:
                # request answer during sync
                rids = msg[:data.MESSAGE_ID_SIZE]
                rid = struct.unpack(data.MESSAGE_ID_FORMAT, rids)[0]
                self.free.add(rid)
                value = msg[data.MESSAGE_ID_SIZE:]
                key = self.reset_requests.get(rid, None)
                if key is None:
                    log.err("Received non-requested RID. Ignoring...")
                else:
                    del self.reset_requests[rid]
                    self.db.set(key, value)
                if (len(self.reset_requests) == 0) and (len(self.to_sync) == 0):
                    # this "if" will only be checked if actionybte == data.ID_ANSWER, which can only be sent during a sync
                    log.msg("Finished sync.")
                    self.sendString(data.ID_SWITCH)
                    self.callback.callback(self)

            else:
                # ignore errors here
                pass

        else:
            log.err("Received unknown Answer. Losing Connection...")
            self.mode = data.MODE_ERROR
            self.callback.errback(Failure(ProtocolError("Unknown Answer!")))
            self.transport.loseConnection()

    def init_reset(self):
        """initiates a database reset."""
        log.msg("Initiating database reset...")
        self.db.reset()
        log.msg("Starting sync...")
        self.sendString(data.ID_SWITCH)
        self.sendString(data.ID_GETKEYS + self.reset_req_string)

    def get_id(self):
        """returns a request id."""
        if len(self.free) > 0:
            return self.free.pop()
        else:
            i = self.cur_id
            if self.cur_id + 1 >= self.range_end:
                raise RuntimeError("No ids left.")
            self.cur_id += 1
            return i



"""The KVNDB Router coordinates the databases and requestss"""
import struct
import random


from twisted.internet.protocol import Factory
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.python.failure import Failure
from twisted.python import log
from twisted.protocols.basic import IntNStringReceiver

from . import data, utils


class RouterFactory(Factory):
    """A twisted.internet.protocol.Factory for the request routing."""
    def __init__(self, reactor, pswd):
        if hasattr(Factory, "__init__"):
            # call __init__ if required
            Factory.__init__(self)
        self.reactor = reactor
        self.pswd = pswd
        self.servers = []
        self.syncing = []
        self.all = []
        self.free_ranges = []
        self.calls = {}

    def buildProtocol(self, addr):
        """builds the protocol."""
        log.msg("Building protocol for connection with client '{a}'...".format(a=addr))
        return RouterProtocol(self)

    def got_answer(self, proto, actionbyte, msg):
        """a server answered a request."""
        rid = struct.unpack(data.MESSAGE_ID_FORMAT, msg[:data.MESSAGE_ID_SIZE])[0]
        if rid not in self.calls:
            # request already answered by another server
            return
        elif actionbyte == data.ID_ANSWER:
            value = msg[data.MESSAGE_ID_SIZE:]
            d = self.calls[rid]
            try:
                d.callback(value)
            finally:
                del self.calls[rid]
        elif actionbyte == data.ID_NOTFOUND:
            d = self.calls[rid]
            fail = Failure(KeyError())
            try:
                d.errback(fail)
            finally:
                del self.calls[rid]
        elif actionbyte == data.ID_ALLKEYS:
            keylist = msg[data.MESSAGE_ID_SIZE:]
            d = self.calls[rid]
            try:
                d.callback(keylist)
            finally:
                del self.calls[rid]
        else:
            raise RuntimeError("Invalid Answer!")

    def get(self, request_id, key):
        """returns the value for key."""
        rid = struct.unpack(data.MESSAGE_ID_FORMAT, request_id)[0]
        d = Deferred()
        if len(self.servers) == 0:
            # no server available; raise KeyError to deferred
            f = Failure(KeyError())
            d.errback(f)
            return d
        self.calls[rid] = d
        server = random.choice(self.servers)
        server.send_get(request_id, key)
        return d

    def set(self, key, value):
        """sets the value for key."""
        for server in self.servers + self.syncing:
            server.send_set(key, value)

    def delete(self, key):
        """deletes the value for key."""
        for server in self.servers + self.syncing:
            server.send_del(key)

    def getkeys(self, request_id,):
        """returns a list of keys."""
        if len(self.servers) == 0:
            # no servers, but client expects sync
            keys = []  # because there are no databases connected.
            answer = utils.keylist2string(keys)
            return answer
        server = random.choice(self.servers)
        rid = struct.unpack(data.MESSAGE_ID_FORMAT, request_id)[0]
        d = Deferred()
        self.calls[rid] = d
        server.send_getkeys(request_id)
        return d

    def get_range(self):
        """returns a rid range for a client."""
        i = 0
        while True:
            start = data.RANGE * i
            end = start + data.RANGE
            found = True
            for client in self.all:
                cs = client.range_start
                if cs == start:
                    found = False
                    break
            if found:
                return (start, end)
            i += 1


class RouterProtocol(IntNStringReceiver):
    """A twisted.internet.protocol.Protocol for the request routing."""
    def __init__(self, factory):
        if hasattr(IntNStringReceiver, "__init__"):
            IntNStringReceiver.__init__(self)
        self.factory = factory
        self.structFormat = data.MESSAGE_LENGTH_FORMAT
        self.prefixLength = data.MESSAGE_LENGTH_SIZE
        self.mode = data.MODE_CONNECTING
        self.can_switch = False
        self.range_start = None
        self.range_end = None

    def connectionMade(self):
        """called when the connection was established."""
        if hasattr(IntNStringReceiver, "connectionMade"):
            IntNStringReceiver.connectionMade(self)
        self.mode = data.MODE_VERSION
        log.msg("Starting Handshake...")

    @inlineCallbacks
    def stringReceived(self, msg):
        """called when a string was received."""
        if self.mode == data.MODE_VERSION:
            # check version
            if len(msg) != data.VERSION_LENGTH:
                log.err("Got invalid version string from client. Aborting connection...")
                # protocol violation
                self.transport.abortConnection()
            version = struct.unpack(data.VERSION_FORMAT, msg)[0]
            if version != data.VERSION:
                # version mismatch
                log.msg("Client uses an unsupported version. Losing connection...")
                self.sendString(data.STATE_ERROR)
                self.transport.loseConnection()
                self.mode = data.MODE_ERROR
            else:
                # proceed with password handling if required
                if self.factory.pswd is not None:
                    log.msg("Version OK, asking for password...")
                    self.sendString(data.STATE_PASSWORD_REQUIRED)
                    self.mode = data.MODE_PASSWORD
                else:
                    log.msg("Version OK, asking for mode...")
                    self.sendString(data.STATE_OK)
                    self.mode = data.MODE_UNKNOWN

        elif self.mode == data.MODE_PASSWORD:
            # check password
            if msg != self.factory.pswd:
                # invalid password
                yield utils.dsleep(self.factory.reactor, data.INVALID_PASSWORD_SLEEP_INTERVAL)  # sleep to prevent high speed attacks.
                log.msg("Client sent invalid password. Losing connection...")
                self.sendString(data.STATE_ERROR)
                self.transport.loseConnection()
                self.mode = data.MODE_ERROR
            else:
                # password correct
                log.msg("Password OK, asking for mode...")
                self.sendString(data.STATE_OK)
                self.mode = data.MODE_UNKNOWN

        elif self.mode == data.MODE_UNKNOWN:
            # handle client mode
            if msg == data.MODE_SERVER:
                # client host db
                log.msg("Client identified as a database. Sending range...")
                self.mode = data.MODE_SERVER
                self.factory.servers.append(self)
                self.can_switch = True
            elif msg == data.MODE_CLIENT:
                # client wants to access database
                log.msg("Client identified as a client. Sending range...")
                self.mode = data.MODE_CLIENT
                self.can_switch = False
            else:
                # protocol violation
                log.err("Client sent invalid mode. Aborting connection...")
                self.transport.abortConnection()

            # now set rid range start and send int
            self.range_start, self.range_end = self.factory.get_range()
            rangestring = struct.pack(data.RANGE_FORMAT, self.range_start, self.range_end)
            self.sendString(rangestring)
            self.factory.all.append(self)

        elif self.mode == data.MODE_SERVER:
            # message from the server
            actionbyte = msg[0]
            if actionbyte in (data.ID_ANSWER, data.ID_NOTFOUND, data.ID_ALLKEYS):
                self.factory.got_answer(self, actionbyte, msg[1:])
            elif actionbyte == data.ID_SWITCH:
                self.switch_mode()

        elif self.mode == data.MODE_CLIENT:
            # request from client
            actionbyte = msg[0]
            if actionbyte == data.ID_SET:
                # set
                keysize = struct.unpack(data.MESSAGE_KEY_LENGTH_FORMAT, msg[1:1 + data.MESSAGE_KEY_LENGTH_SIZE])[0]
                key = msg[1 + data.MESSAGE_KEY_LENGTH_SIZE:1 + data.MESSAGE_KEY_LENGTH_SIZE + keysize]
                value = msg[1 + data.MESSAGE_KEY_LENGTH_SIZE + keysize:]
                self.factory.set(key, value)
            elif actionbyte == data.ID_GET:
                # get
                request_id = msg[1:1 + data.MESSAGE_ID_SIZE]
                key = msg[1 + data.MESSAGE_ID_SIZE:]
                try:
                    value = yield self.factory.get(request_id, key)
                    self.sendString(data.ID_ANSWER + request_id + value)
                except KeyError:
                    self.sendString(data.ID_NOTFOUND + request_id)
            elif actionbyte == data.ID_DEL:
                # del
                key = msg[1:]
                self.factory.delete(key)
            elif actionbyte == data.ID_GETKEYS:
                rid = msg[1:]
                keystring = yield self.factory.getkeys(rid)
                self.sendString(data.ID_ALLKEYS + rid + keystring)
            elif actionbyte == data.ID_SWITCH:
                self.switch_mode()

        else:
            # unknown mode; server side error
            raise RuntimeError("Invalid Mode")

    def connectionLost(self, reason):
        """called when the connection was lost."""
        log.msg("Connection lost. Reason: {r}".format(r=reason))
        if self in self.factory.servers:
            self.factory.servers.remove(self)
        if self in self.factory.syncing:
            self.factory.syncing.remove(self)

        if self in self.factory.all:
            self.factory.all.remove(self)

    def send_get(self, rid, key):
        """if in server mode, request the value for key from the db-server. otherwise, raise AssertionError"""
        assert self.mode == data.MODE_SERVER, "This protocol is not connected to a database!"
        self.sendString(data.ID_GET + rid + key)

    def send_set(self, key, value):
        """if in server mode, tell the db-server to set key to value. otherwise, raise AssertionError"""
        assert self.mode == data.MODE_SERVER or self.can_switch, "This protocol is not connected to a database!"
        keylength = struct.pack(data.MESSAGE_KEY_LENGTH_FORMAT, len(key))
        self.sendString(data.ID_SET + keylength + key + value)

    def send_del(self, key):
        """if in server mode, tell the db-server to delete the value for key. otherwise, raise AssertionError"""
        assert self.mode == data.MODE_SERVER or self.can_switch, "This protocol is not connected to a database!"
        self.sendString(data.ID_DEL + key)

    def send_getkeys(self, rid):
        """if in server mode, requests a list of keys from the db-server. otherwise, raise AssertionError"""
        assert self.mode == data.MODE_SERVER, "This protocol is not connected to a database!"
        self.sendString(data.ID_GETKEYS + rid)

    def switch_mode(self):
        """switch between client/server mode."""
        if self.can_switch:
            if self.mode == data.MODE_SERVER:
                if self in self.factory.servers:
                    self.factory.servers.remove(self)
                if self not in self.factory.syncing:
                    self.factory.syncing.append(self)
                self.mode = data.MODE_CLIENT
            elif self.mode == data.MODE_CLIENT:
                if self not in self.factory.servers:
                    self.factory.servers.append(self)
                if self in self.factory.syncing:
                    self.factory.syncing.remove(self)
                self.mode = data.MODE_SERVER

    # DEBUG FUNCTIONS
    """
    def sendString(self, s):
        print "sending: {s}".format(s=repr(s))
        IntNStringReceiver.sendString(self, s)
    """



#!/usr/bin/env python2
"""The command line interface"""
import sys
import argparse

from twisted.internet import endpoints, reactor
from twisted.python import log

from .ramdb import RamDatabase
from .dbmdb import DbmDatabase
from .dirdb import DirdbmDatabase
from . import router, data, dbproto, cmdclient, txclient

DATABASES = {  # name -> Database
    "ram": RamDatabase,
    "dbm": DbmDatabase,
    "dir": DirdbmDatabase,
}


def run(args=None):
    """runs the command line interface."""
    if args is None:
        args = sys.argv[1:]
    parser = argparse.ArgumentParser(description="Key/Value Network Database")
    parser.add_argument(
        "mode", choices=["router", "cmd"] + DATABASES.keys(), action="store",
        help="What mode the server should operate in",
        )
    parser.add_argument(
        "host", default="0.0.0.0", action="store", nargs="?",
        help="host to serve on/connect to",
        )
    parser.add_argument(
        "port", default=data.DEFAULT_PORT, action="store", type=int, nargs="?",
        help="port to serve on/connect to",
        )
    parser.add_argument(
        "-t", "--type", required=False, default="tcp", action="store", dest="type", nargs="?",
        choices=("tcp", "tcp6", "tls"),
        help="endpoint type to use",
        )
    parser.add_argument(
        "-e", "--endpoint", required=False, default=None, action="store", dest="endpoint",
        help="Use this endpoint instead of creating one from the other arguments",
        )
    parser.add_argument(
        "-p", "--password", action="store", required=False, default=None, dest="password",
        help="password for the server",
        )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", help="print more messages.",
        )
    parser.add_argument(
        "-l", "--logfile", action="store", required=False, dest="logfile", default=None,
        help="file to log to [default: stdout]",
        )
    parser.add_argument(
        "-r", "--reset", action="store_true", required=False, dest="reset",
        help="Reset database and load db from server",
        )
    parser.add_argument(
        "--reset-sleep", action="store", required=False, type=float, default=0.2, dest="sleep_interval",
        help="wait every 128 requests this many seconds before sending the next requests.",
        )
    parser.add_argument(
        "arguments", action="store", nargs="*",
        help="arguments to pass to database interface",
        )
    ns = parser.parse_args(args)

    if ns.verbose:
        if ns.logfile is None:
            log.startLogging(sys.stdout)
        else:
            logfile = open(ns.logfile, "w")
            log.startLogging(logfile)

    if ns.mode == "router":
        if ns.endpoint is None:
            es = "{type}:port={port}:interface={host}".format(type=ns.type, port=ns.port, host=ns.host)
        else:
            es = ns.endpoint
        factory = router.RouterFactory(reactor, ns.password)
        endpoint = endpoints.serverFromString(reactor, es)
        endpoint.listen(factory)
    else:
        if ns.endpoint is None:
            es = "{type}:host={host}:port={port}".format(type=ns.type, port=ns.port, host=ns.host)
        else:
            es = ns.endpoint
        endpoint = endpoints.clientFromString(reactor, es)
        if ns.mode == "cmd":
            cmd = cmdclient.KVNDBCmdClient(reactor)
            protocol = txclient.ClientProtocol(ns.password)
            protocol.callback.addCallback(cmd.defer_entry)
        else:
            db = DATABASES[ns.mode](ns.arguments)
            protocol = dbproto.DatabaseClientProtocol(db, ns.password, reactor, reset=ns.reset, reset_sleep_interval=ns.sleep_interval)
        endpoints.connectProtocol(endpoint, protocol)
    reactor.run()


if __name__ == "__main__":
    run(sys.argv[1:])



"""the setup script for KVNDB."""
from setuptools import setup


def read_file(path):
    with open(path, "r") as f:
        return f.read()


setup(
    name="kvndb",
    version="0.1",
    description="A modular Key/Value Network Database server+client using twisted",
    long_description=read_file("README.md"),
    author="bennr01",
    author_email="benjamin99.vogt@web.de",
    url="https://github.com/bennr01/kvndb/",
    license="MIT",
    packages=["kvndb"],
    zip_safe=False,
    install_requires=[
        "twisted",
        ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2.7",
        "Topic :: Database",
        ],
    keywords="database key value network scalable",
    entry_points={
        "console_scripts": [
            "kvndb=kvndb.runner:run",
            ],
        },
    )



"""Command line client interface."""
import cmd
import shlex

from twisted.internet import threads

from . import data, utils


class KVNDBCmdClient(cmd.Cmd):
    """A command line interface for a KVNDB client."""
    
    prompt = "(?)"
    intro = "KVNDB Cmd-Client Protocol v{v}\nType 'help' for help.".format(v=data.VERSION)

    def __init__(self, reactor, *args, **kwargs):
        cmd.Cmd.__init__(self, *args, **kwargs)
        self.reactor = reactor
        self.proto = None

    def do_set(self, cmd):
        """set KEY VALUE: set KEY to VALUE."""
        args = shlex.split(cmd)
        if len(args) != 2:
            self.stdout.write("Error: Expected 2 arguments, got {n} instead!\n".format(n=len(args)))
            return
        else:
            key, value = args
            self.reactor.callFromThread(self.proto.set, key, value)

    def do_del(self, cmd):
        """del KEY: delete KEY"""
        self.reactor.callFromThread(self.proto.delete, cmd)

    do_delete = do_remove = do_del

    def do_get(self, cmd):
        """get KEY: show the value of KEY"""
        key = cmd
        try:
            value = threads.blockingCallFromThread(
                self.reactor, self.proto.get, key,
                )
        except KeyError:
            self.stdout.write("Error: Key '{k}' not found!\n".format(k=key))
            return
        except Exception as e:
            self.stdout.write("Error: {e}\n".format(e=e))
            return
        else:
            self.stdout.write("{v}\n".format(v=value))

    do_show = do_view = do_print = do_get

    def do_getkeys(self, cmd):
        """getkeys [COLUMNS]: show a list of all keys."""
        if cmd:
            try:
                cols = int(cmd)
            except ValueError:
                self.stdout.write("Error: Invalid argument!\n")
                return
        else:
            cols = 2
        try:
            keys = threads.blockingCallFromThread(
                self.reactor, self.proto.getkeys,
                )
        except Exception as e:
            self.stdout.write("Error: {e}\n".format(e=e))
            return
        else:
            sk = sorted(keys)
            formated = utils.fmtcols(sk, cols)
            if not formated.endswith("\n"):
                formated += "\n"
            self.stdout.write(formated)

    do_keys = do_list = do_getkeys

    def defer_entry(self, proto):
        """
this is a utility function which should be used by the ClientProtocol.
When called, the protocol will be set and the REPL-starts.
"""
        self.proto = proto
        self.reactor.callInThread(self.cmdloop)

    def do_EOF(self, cmd=""):
        """EOF: quit the command loop."""
        self.reactor.callFromThread(self.reactor.stop)
        return True

    do_quit = do_exit = do_EOF

    def postcmd(self, stop, line):
        """called after a line has been executed."""
        return stop
